//
// This code is part of GrapeCity.Documents.Word demos.
// Copyright (c) GrapeCity, Inc. All rights reserved.
//
// This code uses the Markdig library by Alexandre Mutel,
// see https://github.com/xoofx/markdig.
//
using System;
using System.Text.RegularExpressions;
using System.Linq;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using Markdig.Syntax;
using GrapeCity.Documents.Word;

namespace DsWordWeb.Demos.MarkdownToWordRenderer
{
    /// <summary>
    /// An Word renderer for a <see cref="HeadingBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{HeadingBlock}" />
    public class HeadingRenderer : WordObjectRenderer<HeadingBlock>
    {
        protected override void Write(WordRenderer renderer, HeadingBlock obj)
        {
            renderer.InHtmlBlock = false;
            int index = obj.Level - 1;
            Style style;
            if (index >= 0 && index < renderer.HeaderStyles.Length)
                style = renderer.HeaderStyles[index];
            else
                style = renderer.HeaderFallbackStyle;

            if (renderer.IsCurrBlockStyleDefault)
            {
                var prevBlockStyle = renderer.CurrBlockStyle;
                renderer.CurrBlockStyle = style;
                renderer.AddParagraph();
                renderer.WriteLeafInline(obj);
                renderer.CurrBlockStyle = prevBlockStyle;
            }
            else
            {
                var prevInlineStyle = renderer.CurrInlineStyle;
                var myInlineStyleProps = WordRenderer.InlineStyleProps.FromStyle(style);
                var myInlineStyle = myInlineStyleProps.MergeWith(prevInlineStyle);
                renderer.CurrInlineStyle = myInlineStyle;
                renderer.AddParagraph();
                renderer.WriteLeafInline(obj);
                renderer.CurrInlineStyle = prevInlineStyle;
            }
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="ParagraphBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{ParagraphBlock}" />
    public class ParagraphRenderer : WordObjectRenderer<ParagraphBlock>
    {
        protected override void Write(WordRenderer renderer, ParagraphBlock obj)
        {
            renderer.AddParagraph();
            renderer.InHtmlBlock = false;
            renderer.WriteLeafInline(obj);
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="ListBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{ListBlock}" />
    public class ListRenderer : WordObjectRenderer<ListBlock>
    {
        protected override void Write(WordRenderer renderer, ListBlock listBlock)
        {
            renderer.InHtmlBlock = false;
            if (listBlock.IsOrdered)
            {
                // TODO: use listBlock.OrderedStart for RestartAt:
                renderer.ListStack.Push(new WordRenderer.ListItemProps()
                    {
                        RestartAt = 1,
                        Template = renderer.NumberedListTemplate,
                        Level = renderer.ListStack.Count,
                    }
                );
            }
            else
            {
                renderer.ListStack.Push(new WordRenderer.ListItemProps()
                    {
                        RestartAt = -1,
                        Template = renderer.BulletListTemplate,
                        Level = renderer.ListStack.Count,
                    }
                );
            }
            foreach (var item in listBlock)
            {
                var listItem = (ListItemBlock)item;
                renderer.WriteChildren(listItem);
                renderer.ListStack.Peek().IsContinuation = false;
            }
            renderer.ListStack.Pop();
        }
    }

    /// <summary>
    /// An Word renderer for a <see cref="CodeBlock"/> and <see cref="FencedCodeBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{CodeBlock}" />
    public class CodeBlockRenderer : WordObjectRenderer<CodeBlock>
    {
        protected override void Write(WordRenderer renderer, CodeBlock obj)
        {
            renderer.InHtmlBlock = false;
            var prevBlockStyle = renderer.CurrBlockStyle;
            renderer.CurrBlockStyle = renderer.CodeBlockStyle;
            renderer.AddParagraph();
            renderer.WriteCodeLines(obj);
            renderer.CurrBlockStyle = renderer.CodeBlockBreakStyle;
            renderer.AddParagraph();
            renderer.CurrBlockStyle = prevBlockStyle;
        }
    }

    public class ThematicBreakRenderer : WordObjectRenderer<ThematicBreakBlock>
    {
        protected override void Write(WordRenderer renderer, ThematicBreakBlock obj)
        {
            renderer.InHtmlBlock = false;
            var prevBlockStyle = renderer.CurrBlockStyle;
            renderer.CurrBlockStyle = renderer.ThematicBreakStyle;
            renderer.AddParagraph();
            renderer.CurrBlockStyle = prevBlockStyle;
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="QuoteBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{QuoteBlock}" />
    public class QuoteBlockRenderer : WordObjectRenderer<QuoteBlock>
    {
        protected override void Write(WordRenderer renderer, QuoteBlock obj)
        {
            if (!renderer.IsCurrBlockStyleDefault)
                throw new InvalidOperationException("Current block style should be default when starting quote block.");

            renderer.InHtmlBlock = false;
            var prevBlockStyle = renderer.CurrBlockStyle;
            renderer.CurrBlockStyle = renderer.QuoteBlockStyle;
            renderer.WriteChildren(obj);
            renderer.CurrBlockStyle = prevBlockStyle;
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="HtmlBlock"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{HtmlBlock}" />
    public class HtmlBlockRenderer : WordObjectRenderer<HtmlBlock>
    {
        private static readonly HashSet<string> s_blocks = new HashSet<string>() {
            "address", "article", "aside", "blockquote", "canvas", "dd", "div", "dl", "dt",
            "fieldset", "figcaption", "figure", "footer", "form", "header",
            "h1", "h2", "h3", "h4", "h5", "h6", "hr", "li",  "main", "nav", "noscript",
            "ol", "p", "pre", "section", "table", "tfoot", "ul", "video",
        };
        private static readonly Regex s_hdr = new Regex(@"\<(\/?)([a-z]+)([\d]?)\s?[^>]*>(.*)", RegexOptions.Compiled | RegexOptions.Singleline);

        internal static bool CheckKnownHtmlBlock(WordRenderer renderer, string txt, out bool closingTag, out string value)
        {
            var m = s_hdr.Match(txt);
            if (!m.Success)
            {
                if (txt.StartsWith("<!--") && txt.EndsWith("-->"))
                {
                    closingTag = false;
                    value = null;
                    return false;
                }
                throw new ArgumentException($"Invalid html tag: {txt}");
            }

            // Groups:
            // 1 - null or "/"
            // 2 - tag w/out digit suffix
            // 3 - null or digit suffix
            // 4 - text after closing ">"
            closingTag = m.Groups[1].Value == "/";
            if (m.Groups[2].Value == "h")
            {
                var tag = m.Groups[2].Value + m.Groups[3].Value;
                ProcessBlockTag(renderer, tag, closingTag);
                value = closingTag ? null : m.Groups[4].Value;
                return true;
            }
            else
            {
                closingTag = m.Groups[1].Value == "/";
                if (!closingTag)
                {
                    value = m.Groups[4].Value;
                    return s_blocks.Contains(m.Groups[2].Value);
                }
                else
                {
                    value = null;
                    return false;
                }
            }
        }

        private static bool ProcessBlockTag(WordRenderer renderer, string tag, bool closingTag)
        {
            Debug.Assert(s_blocks.Contains(tag));
            switch (tag)
            {
                case "h1":
                case "h2":
                case "h3":
                case "h4":
                case "h5":
                case "h6":
                    if (!closingTag)
                    {
                        Style style;
                        var index = tag[1] - '1';
                        if (index >= 0 && index < renderer.HeaderStyles.Length)
                            style = renderer.HeaderStyles[index];
                        else
                            style = renderer.HeaderFallbackStyle;
                        renderer.HtmlBlockStack.Push((tag, renderer.CurrBlockStyle));
                        renderer.CurrBlockStyle = style;
                    }
                    else
                    {
                        var top = renderer.HtmlBlockStack.Pop();
                        renderer.CurrBlockStyle = top.style;
                    }
                    return true;
                default:
                    // TODO: handle other tags such as <code>.
                    return false;
            }
        }

        private static readonly Regex s_htags = new Regex(
            @"\<(?<end>/?)(?<tag>h[1-6]|[a-z]+)[^>]*\>(?<val>[^<]*)|(?<comment>\<!--.+?--\>)|(?<rest>[^<]+)",
            RegexOptions.Compiled | RegexOptions.Singleline);

        protected override void Write(WordRenderer renderer, HtmlBlock obj)
        {
            renderer.InHtmlBlock = true;

            // TODO: optimize?
            var txt = obj.Lines.ToString().TrimStart();

            var matches = s_htags.Matches(txt);
            foreach (Match m in matches)
            {
                Debug.Assert(m.Success);
                // Console.WriteLine($"HTML: {m.Groups["val"].Value}");

                var value = m.Groups["val"].Value;
                value = WebUtility.HtmlDecode(value);
                // TODO: consider behavior inside <pre> tag:
                var tag = m.Groups["tag"].Value;
                var closing = m.Groups["end"].Value == "/";
                if (s_blocks.Contains(tag))
                {
                    // TODO: not sure what to do in this case, add newline or just a paragraph:
                    Debug.Assert(value != "\r\n");
                    ProcessBlockTag(renderer, tag, closing);
                    if (!closing)
                        renderer.AddParagraph(value);
                }
                else
                    renderer.AddRun(value);
            }
        }
    }
}
