//
// This code is part of GrapeCity.Documents.Word demos.
// Copyright (c) GrapeCity, Inc. All rights reserved.
//
// This code uses the Markdig library by Alexandre Mutel,
// see https://github.com/xoofx/markdig.
//
using System;
using System.Linq;
using System.Drawing;
using System.Collections.Generic;
using System.Net.Mail;
using System.Xml.Linq;
using Markdig.Syntax.Inlines;
using GrapeCity.Documents.Word;

namespace DsWordWeb.Demos.MarkdownToWordRenderer
{
    /// <summary>
    /// A Word renderer for a <see cref="LiteralInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{LiteralInline}" />
    public class LiteralInlineRenderer : WordObjectRenderer<LiteralInline>
    {
        protected override void Write(WordRenderer renderer, LiteralInline obj)
        {
            renderer.Write(ref obj.Content);
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="CodeInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{CodeInline}" />
    public class CodeInlineRenderer : WordObjectRenderer<CodeInline>
    {
        protected override void Write(WordRenderer renderer, CodeInline obj)
        {
            var prevInlineStyle = renderer.CurrInlineStyle;
            var myInlineStyleProps = WordRenderer.InlineStyleProps.FromStyle(renderer.CodeInlineStyle);
            var myInlineStyle = myInlineStyleProps.MergeWith(prevInlineStyle);

            renderer.CurrInlineStyle = myInlineStyle;
            renderer.Write(obj.ContentSpan);
            renderer.CurrInlineStyle = prevInlineStyle;
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="EmphasisInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{EmphasisInline}" />
    public class EmphasisInlineRenderer : WordObjectRenderer<EmphasisInline>
    {
        protected override void Write(WordRenderer renderer, EmphasisInline obj)
        {
            var myInlineStyleProps = new WordRenderer.InlineStyleProps();
            string unknown = null;
            switch (obj.DelimiterChar)
            {
                case '~':
                    if (obj.DelimiterCount == 2)
                        myInlineStyleProps.StrikeThrough = true;
                    else
                        goto default;
                    break;
                case '*':
                case '_':
                    if (obj.DelimiterCount == 2)
                        myInlineStyleProps.Bold = true;
                    else // obj.DelimiterCount == 1
                        myInlineStyleProps.Italic = true;
                    break;
                default:
                    // We do not handle extensions:
                    //  ~Subscript~
                    //  ^Superscript^
                    //  ++Inserted++
                    //  ==Marked==
                    unknown = new string(obj.DelimiterChar, obj.DelimiterCount);
                    break;
            }

            var prevInlineStyle = renderer.CurrInlineStyle;
            var myInlineStyle = myInlineStyleProps.MergeWith(prevInlineStyle);

            if (unknown != null)
                renderer.Write(unknown);
            else if (myInlineStyle != prevInlineStyle)
                renderer.CurrInlineStyle = myInlineStyle;

            renderer.WriteChildren(obj);

            if (unknown != null)
                renderer.Write(unknown);
            else if (myInlineStyle != prevInlineStyle)
                renderer.CurrInlineStyle = prevInlineStyle;
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="LineBreakInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{LineBreakInline}" />
    public class LineBreakInlineRenderer : WordObjectRenderer<LineBreakInline>
    {
        protected override void Write(WordRenderer renderer, LineBreakInline obj)
        {
            if (renderer.IsLastInContainer)
                return;

            if (obj.IsHard)
                renderer.AddBreak();
            else
                renderer.AddRun(" ");
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="DelimiterInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{DelimiterInline}" />
    public class DelimiterInlineRenderer : WordObjectRenderer<DelimiterInline>
    {
        protected override void Write(WordRenderer renderer, DelimiterInline obj)
        {
            renderer.Write(obj.ToLiteral());
            renderer.WriteChildren(obj);
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="LinkInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{LinkInline}" />
    public class LinkInlineRenderer : WordObjectRenderer<LinkInline>
    {
        protected override void Write(WordRenderer renderer, LinkInline link)
        {
            // TODO: handle images (link.IsImage)
            // renderer.Write(link.IsImage ? "<img src=\"" : "<a href=\"");
            var url = link.GetDynamicUrl != null ? link.GetDynamicUrl() ?? link.Url : link.Url;

            var a = renderer.AddHyperlink();
            if (url.StartsWith('#'))
                a.Anchor = url[1..];
            else
            {
#if MailAddress_TryCreate
                if (MailAddress.TryCreate(url, out MailAddress email))
                    url = $"mailto:{email.Address}";
#else
                try
                {
                    // TODO: MailAddress ctor is SLOW, maybe do w/out it?
                    var email = new MailAddress(url);
                    url = $"mailto:{email.Address}";
                }
                catch
                {
                    // Ignore errors.
                }
#endif
                a.Address = new Uri(url);
                if (!string.IsNullOrEmpty(link.Title))
                    a.ScreenTip = link.Title;
            }

            var prevInlineStyle = renderer.CurrInlineStyle;
            var myInlineStyleProps = WordRenderer.InlineStyleProps.FromStyle(renderer.LinkInlineStyle);
            var myInlineStyle = myInlineStyleProps.MergeWith(prevInlineStyle);
            var prevCurrHyperlink = renderer.CurrHyperlink;

            renderer.CurrInlineStyle = myInlineStyle;
            renderer.CurrHyperlink = a;
            renderer.WriteChildren(link);
            renderer.CurrHyperlink = prevCurrHyperlink;
            renderer.CurrInlineStyle = prevInlineStyle;
        }
    }

    /// <summary>
    /// A Word renderer for an <see cref="AutolinkInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{AutolinkInline}" />
    public class AutolinkInlineRenderer : WordObjectRenderer<AutolinkInline>
    {
        protected override void Write(WordRenderer renderer, AutolinkInline obj)
        {
            var url = obj.Url;
            if (obj.IsEmail)
                url = $"mailto:{url}";

            var a = renderer.AddHyperlink();
            a.Address = new Uri(url);

            var prevInlineStyle = renderer.CurrInlineStyle;
            var myInlineStyleProps = WordRenderer.InlineStyleProps.FromStyle(renderer.LinkInlineStyle);
            var myInlineStyle = myInlineStyleProps.MergeWith(prevInlineStyle);
            var prevCurrHyperlink = renderer.CurrHyperlink;

            renderer.CurrInlineStyle = myInlineStyle;
            renderer.CurrHyperlink = a;
            renderer.Write(obj.Url);
            renderer.CurrHyperlink = prevCurrHyperlink;
            renderer.CurrInlineStyle = prevInlineStyle;
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="HtmlInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{HtmlInline}" />
    public class HtmlInlineRenderer : WordObjectRenderer<HtmlInline>
    {
        protected override void Write(WordRenderer renderer, HtmlInline obj)
        {
            var tag = obj.Tag ?? string.Empty;
            if (tag.StartsWith("<a "))
            {
                string bookmark;
                tag += "</a>";
                var a = XElement.Parse(tag);
                if (a.Attribute("href")?.Value is string href)
                    bookmark = href[1..];
                else if (a.Attribute("id")?.Value is string id)
                    bookmark = id;
                else
                    bookmark = null;
                if (bookmark != null)
                    renderer.AddBookmark(bookmark);
            }
            else
            {
                var known = HtmlBlockRenderer.CheckKnownHtmlBlock(renderer, tag, out bool closingTag, out string value);
                if (known)
                    renderer.AddParagraph();
                else if (!closingTag)
                {
                    var i = tag.IndexOfAny(new char[] { ' ', '>' });
                    if (i < 0)
                        throw new Exception("Invalid html tag.");
                    tag += "</" + tag[1..(i + 1)];
                    if (tag[i] != '>')
                        tag += ">";
                    var e = XElement.Parse(tag);
                    renderer.AddRun(e.Value);
                }
            }
        }
    }

    /// <summary>
    /// A Word renderer for a <see cref="HtmlEntityInline"/>.
    /// </summary>
    /// <seealso cref="WordObjectRenderer{HtmlEntityInline}" />
    public class HtmlEntityInlineRenderer : WordObjectRenderer<HtmlEntityInline>
    {
        protected override void Write(WordRenderer renderer, HtmlEntityInline obj)
        {
            renderer.Write(obj.Transcoded);
        }
    }
}
