//
// This code is part of GrapeCity.Documents.Word demos.
// Copyright (c) GrapeCity, Inc. All rights reserved.
//
// This code uses the Markdig library by Alexandre Mutel,
// see https://github.com/xoofx/markdig.
//
#nullable disable
using System;
using System.Globalization;
using System.IO;
using System.Drawing;
using System.Runtime.CompilerServices;
using System.Text;
using System.Collections.Generic;
using Markdig;
using Markdig.Helpers;
using Markdig.Syntax.Inlines;
using Markdig.Syntax;
using Markdig.Renderers;
using GrapeCity.Documents.Word;

namespace DsWordWeb.Demos.MarkdownToWordRenderer
{
    /// <summary>
    /// A base class for rendering <see cref="Block"/> and <see cref="Inline"/> Markdown objects to Word.
    /// </summary>
    /// <typeparam name="TObject">The type of the object.</typeparam>
    /// <seealso cref="IMarkdownObjectRenderer" />
    public abstract class WordObjectRenderer<TObject> : MarkdownObjectRenderer<WordRenderer, TObject> where TObject : MarkdownObject
    {
    }

    /// <summary>
    /// Represents a renderer that can be used to convert
    /// a Markdown document to a Word document.
    /// </summary>
    /// <seealso cref="TextRendererBase{WordRenderer}" />
    public class WordRenderer : RendererBase
    {
        private static readonly MarkdownPipeline _defaultPipeline = new MarkdownPipelineBuilder().Build();

        /// <summary>
        /// Converts a Markdown string to a <see cref="GcWordDocument"/> object.
        /// </summary>
        /// <param name="markdown">The markdown string</param>
        /// <param name="pipeline">The pipeline used for parsing the markdown (see Markdig documentation for details).</param>
        /// <returns>The Word document representing the markdown.</returns>
        public static GcWordDocument ToWord(string markdown, MarkdownPipeline pipeline = null)
        {
            pipeline = pipeline ?? _defaultPipeline;
            var mdoc = Markdig.Markdown.Parse(markdown, pipeline);
            return ToWord(mdoc);
        }

        /// <summary>
        /// Converts a <see cref="MarkdownDocument"/> to a <see cref="GcWordDocument"/> object.
        /// </summary>
        /// <param name="document">The markdown object.</param>
        /// <returns>The Word document representing the markdown.</returns>
        public static GcWordDocument ToWord(MarkdownDocument document)
        {
            var doc = new GcWordDocument();
            doc.Body.Sections.First.PageSetup.Margin.Left = 36;
            doc.Body.Sections.First.PageSetup.Margin.Top = 36;
            doc.Body.Sections.First.PageSetup.Margin.Right = 36;
            doc.Body.Sections.First.PageSetup.Margin.Bottom = 36;
            var renderer = new WordRenderer(doc);
            renderer.Render(document);
            return doc;
        }

        private GcWordDocument _doc;
        private Style _defaultBlockStyle;

        public Style[] HeaderStyles { get; }
        public Style HeaderFallbackStyle { get; set; }
        public GcWordDocument Document => _doc;
        public Paragraph LastParagraph => _doc.Body.Paragraphs.Last;
        public Style CodeBlockStyle { get; set; }
        public Style CodeInlineStyle { get; set; }
        public Style CodeBlockBreakStyle { get; set; }
        public Style ThematicBreakStyle { get; set; }
        public Style QuoteBlockStyle { get; private set; }
        public Style LinkInlineStyle { get; private set; }
        public Style CurrBlockStyle { get; set; }
        public Style CurrInlineStyle { get; set; }
        public Hyperlink CurrHyperlink { get; set; }
        public ListTemplate BulletListTemplate { get; private set; }
        public ListTemplate NumberedListTemplate { get; private set; }
        public Style[] BulletListContStyles { get; private set; }
        public Style[] NumberedListContStyles { get; private set; }

        public class ListItemProps
        {
            public int RestartAt { get; set; } = -1;
            public int Level { get; set; } = -1;
            public ListTemplate Template { get; set; } = null;
            public bool IsContinuation { get; set; } = false;
        }
        public Stack<ListItemProps> ListStack { get; } = new Stack<ListItemProps>();

        public Stack<(string html, Style style)> HtmlBlockStack { get; } = new Stack<(string, Style)>();

        public bool InHtmlBlock { get; set; } = false;
        /// <summary>
        /// Indicates that last added was empty paragraph inside an HTML block.
        /// </summary>
        private bool IsLastHtmlNewline { get; set; } = false;

        public class InlineStyleProps
        {
            public static readonly InlineStyleProps Default = new InlineStyleProps()
            {
                Bold = false,
                Italic = false,
                StrikeThrough = false,
                FontName = "Calibri",
                FontSize = 11,
                FontColor = Color.Black,
                BackColor = Color.White,
            };

            public bool? Bold { get; set; } = null;
            public bool? Italic { get; set; } = null;
            public bool? StrikeThrough { get; set; } = null;
            public string FontName { get; set; } = null;
            public float? FontSize { get; set; } = null;
            public Color? FontColor { get; set; } = null;
            public Color? BackColor { get; set; } = null;
            static public InlineStyleProps FromStyle(Style style)
            {
                // GcWordDocument.GetPropertyValueSource(() => doc.Body.Runs.First.Font.Size);
                return new InlineStyleProps()
                {
                    Bold = style.Font.Bold,
                    Italic = style.Font.Italic,
                    StrikeThrough = style.Font.StrikeThrough,
                    FontName = style.Font.Name,
                    FontSize = style.Font.Size,
                    FontColor = style.Font.Color.RGB,
                    BackColor = style.Font.Shading.BackgroundPatternColor.RGB,
                };
            }

            public string MakeName()
            {
                return
                    $"md_Inline-" +
                    $"{FontName ?? "_"}-" +
                    $"{FontSize?.ToString()[0] ?? '_'}-" +
                    $"{Bold?.ToString()[0] ?? '_'}-" +
                    $"{Italic?.ToString()[0] ?? '_'}-" +
                    $"{StrikeThrough?.ToString()[0] ?? '_'}-" +
                    $"{FontColor?.ToArgb().ToString("X") ?? "_"}-" +
                    $"{BackColor?.ToArgb().ToString("X") ?? "_"}-" +
                    "";
            }

            public bool IsMergeNeeded(Style other)
            {
                return (Bold.HasValue && Bold.Value != other.Font.Bold) ||
                    (Italic.HasValue && Italic.Value != other.Font.Italic) ||
                    (StrikeThrough.HasValue && StrikeThrough.Value != other.Font.StrikeThrough) ||
                    (FontName != null && FontName != other.Font.Name) ||
                    (FontSize.HasValue && FontSize.Value != other.Font.Size) ||
                    (FontColor.HasValue && FontColor.Value != other.Font.Color.RGB) ||
                    (BackColor.HasValue && BackColor.Value != other.Font.Shading.BackgroundPatternColor.RGB)
                    ;
            }

            public Style MergeWith(Style style)
            {
                if (!IsMergeNeeded(style))
                    return style;

                var sname = MakeName();
                if (style.Document.Styles.TryGetValue(sname, out Style res))
                {
                    return res;
                }
                res = style.Document.Styles.Add(MakeName(), StyleType.Character);
                res.BaseStyle = style;
                if (Bold.HasValue)
                    res.Font.Bold = Bold.Value;
                if (Italic.HasValue)
                    res.Font.Italic = Italic.Value;
                if (StrikeThrough.HasValue)
                    res.Font.StrikeThrough = StrikeThrough.Value;
                if (FontName != null)
                    res.Font.Name = FontName;
                if (FontSize.HasValue)
                    res.Font.Size = FontSize.Value;
                if (FontColor.HasValue)
                    res.Font.Color.RGB = FontColor.Value;
                if (BackColor.HasValue)
                    res.Font.Shading.BackgroundPatternColor.RGB = BackColor.Value;
                return res;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WordRenderer"/> class.
        /// </summary>
        /// <param name="writer">The writer.</param>
        public WordRenderer(GcWordDocument doc) : base()
        {
            _doc = doc;

            // Headers:
            var headerStyleIds = new BuiltInStyleId[]
            {
                BuiltInStyleId.Heading1,
                BuiltInStyleId.Heading2,
                BuiltInStyleId.Heading3,
                BuiltInStyleId.Heading4,
                BuiltInStyleId.Heading5,
                BuiltInStyleId.Heading6,
            };
            HeaderStyles = new Style[headerStyleIds.Length];
            doc.Styles[BuiltInStyleId.Heading1].Font.Size = 18;
            doc.Styles[BuiltInStyleId.Heading1].ParagraphFormat.Borders.Bottom.LineWidth = 0.5f;
            doc.Styles[BuiltInStyleId.Heading1].ParagraphFormat.Borders.Bottom.Color.RGB = Color.LightGray;
            doc.Styles[BuiltInStyleId.Heading1].ParagraphFormat.Spacing.SpaceAfter = 12;
            doc.Styles[BuiltInStyleId.Heading2].Font.Size = 16;
            doc.Styles[BuiltInStyleId.Heading2].ParagraphFormat.Borders.Bottom.LineWidth = 0.5f;
            doc.Styles[BuiltInStyleId.Heading2].ParagraphFormat.Borders.Bottom.Color.RGB = Color.LightGray;
            doc.Styles[BuiltInStyleId.Heading2].ParagraphFormat.Spacing.SpaceAfter = 8;
            doc.Styles[BuiltInStyleId.Heading3].Font.Size = 14;
            doc.Styles[BuiltInStyleId.Heading4].Font.Size = 13;
            doc.Styles[BuiltInStyleId.Heading5].Font.Size = 12;
            doc.Styles[BuiltInStyleId.Heading6].Font.Size = 11;
            for (int i = 0; i < headerStyleIds.Length; ++i)
            {
                var style = HeaderStyles[i] = doc.Styles[headerStyleIds[i]];
                style.Font.Color.RGB = Color.Black;
                style.Font.Bold = true;
                style.Font.Italic = false;
            }
            HeaderFallbackStyle = doc.Styles[BuiltInStyleId.Header];
            // Default block style:
            CurrBlockStyle = _defaultBlockStyle = doc.Styles.Add("md_Block", StyleType.Paragraph);
            // Default inline style:
            // CurrInlineStyle = InlineStyleProps.Default.MergeWith(doc.Styles.Add("md_Inline", StyleType.Character));
            CurrInlineStyle = doc.Styles.Add("md_Inline", StyleType.Character);

            // Code blocks:
            CodeBlockStyle = doc.Styles.Add("md_CodeBlock", StyleType.Paragraph);
            foreach (var b in CodeBlockStyle.ParagraphFormat.Borders)
            {
                b.LineWidth = 0.5f;
                b.Color.RGB = Color.LightGray;
                b.Space = 3;
            }
            CodeBlockStyle.Font.Name = "Consolas";// "Courier New";
            CodeBlockStyle.Font.Size = 9;
            // Format for empty paragraphs after code blocks so that borders of consecutive blocks do not merge:
            CodeBlockBreakStyle = doc.Styles.Add("md_CodeBlockBreak", StyleType.Paragraph);
            CodeBlockBreakStyle.Font.Size = 1;
            foreach (var b in CodeBlockBreakStyle.ParagraphFormat.Borders)
                b.Space = 0;
            CodeBlockBreakStyle.ParagraphFormat.Spacing.SpaceAfterAuto = false;
            CodeBlockBreakStyle.ParagraphFormat.Spacing.SpaceBefore = 0;
            CodeBlockBreakStyle.ParagraphFormat.Spacing.SpaceAfter = 0;
            CodeBlockBreakStyle.ParagraphFormat.Spacing.LineUnitAfter = 0;
            CodeBlockBreakStyle.ParagraphFormat.Spacing.LineSpacing = 0;
            // Code inline:
            CodeInlineStyle = doc.Styles.Add("md_CodeInline", StyleType.Character);
            CodeInlineStyle.Font.Name = "Consolas";
            CodeInlineStyle.Font.Size = 9;
            CodeInlineStyle.Font.Color.RGB = Color.Crimson;
            CodeInlineStyle.Font.Shading.BackgroundPatternColor.RGB = Color.FromArgb(unchecked((int)0xFFF1F1F1));
            // Thematic break (horizontal line):
            ThematicBreakStyle = doc.Styles.Add("md_ThematicBreak", StyleType.Paragraph);
            ThematicBreakStyle.ParagraphFormat.Borders.Bottom.LineWidth = 0.5f;
            ThematicBreakStyle.ParagraphFormat.Borders.Bottom.Color.RGB = Color.LightGray;
            ThematicBreakStyle.ParagraphFormat.Spacing.SpaceAfter = 8;
            // Block quote:
            QuoteBlockStyle = doc.Styles.Add("md_QuoteBlock", StyleType.Paragraph);
            QuoteBlockStyle.Font.Color.RGB = Color.DarkGray;
            QuoteBlockStyle.ParagraphFormat.Indentation.LeftIndent = 16;
            QuoteBlockStyle.ParagraphFormat.Borders.Left.LineWidth = 2f;
            QuoteBlockStyle.ParagraphFormat.Borders.Left.Color.RGB = Color.LightGray;
            QuoteBlockStyle.ParagraphFormat.Borders.Left.Space = 16;
            // Inline links:
            LinkInlineStyle = doc.Styles.Add("md_Link", StyleType.Character);
            LinkInlineStyle.Font.Color.RGB = Color.FromArgb(unchecked((int)0xFF0000EE));

            // Base list templates:
            BulletListTemplate = doc.ListTemplates.Add(BuiltInListTemplateId.BulletDefault, "md_BulletListTemplate");
            BulletListContStyles = new Style[BulletListTemplate.ListLevels.Count];
            for (int i = 0; i < BulletListTemplate.ListLevels.Count; ++i)
            {
                var l = BulletListTemplate.ListLevels[i];
                l.ParagraphFormat.Indentation.LeftIndent -= 16;
                l.NumberPosition += 4;
                var s = BulletListContStyles[i] = _doc.Styles.Add($"md_blCont{i}", StyleType.Paragraph);
                s.BaseStyle = _defaultBlockStyle;
                s.ParagraphFormat.Indentation.LeftIndent = l.ParagraphFormat.Indentation.LeftIndent;
            }
            NumberedListTemplate = doc.ListTemplates.Add(BuiltInListTemplateId.NumberDefault, "md_NumberedListTemplate");
            NumberedListContStyles = new Style[NumberedListTemplate.ListLevels.Count];
            for (int i = 0; i < NumberedListTemplate.ListLevels.Count; ++i)
            {
                var l = NumberedListTemplate.ListLevels[i];
                l.ParagraphFormat.Indentation.LeftIndent -= 16;
                l.NumberPosition += 4;
                var s = NumberedListContStyles[i] = _doc.Styles.Add($"md_nlCont{i}", StyleType.Paragraph);
                s.BaseStyle = _defaultBlockStyle;
                s.ParagraphFormat.Indentation.LeftIndent = l.ParagraphFormat.Indentation.LeftIndent;
            }

            // Block renderers
            ObjectRenderers.Add(new HeadingRenderer());
            ObjectRenderers.Add(new ParagraphRenderer());
            ObjectRenderers.Add(new ListRenderer());
            ObjectRenderers.Add(new CodeBlockRenderer());
            ObjectRenderers.Add(new ThematicBreakRenderer());
            ObjectRenderers.Add(new QuoteBlockRenderer());
            ObjectRenderers.Add(new HtmlBlockRenderer());

            // Inline renderers
            ObjectRenderers.Add(new LiteralInlineRenderer());
            ObjectRenderers.Add(new CodeInlineRenderer());
            ObjectRenderers.Add(new EmphasisInlineRenderer());
            ObjectRenderers.Add(new LineBreakInlineRenderer());
            ObjectRenderers.Add(new DelimiterInlineRenderer());
            ObjectRenderers.Add(new LinkInlineRenderer());
            ObjectRenderers.Add(new AutolinkInlineRenderer());
            ObjectRenderers.Add(new HtmlInlineRenderer());
            ObjectRenderers.Add(new HtmlEntityInlineRenderer());
        }

        public override object Render(MarkdownObject markdownObject)
        {
            Write(markdownObject);
            return _doc;
        }

        public bool IsCurrBlockStyleDefault => CurrBlockStyle == _defaultBlockStyle;

        public Paragraph AddParagraph(string text = null)
        {
            var thisIsNewline = string.IsNullOrEmpty(text);
            if (IsLastHtmlNewline && thisIsNewline)
            {
                _doc.Body.Paragraphs.Last.Style = CurrBlockStyle;
                return _doc.Body.Paragraphs.Last;
            }
            IsLastHtmlNewline = thisIsNewline && InHtmlBlock;

            Paragraph p;
            if (thisIsNewline)
                p = _doc.Body.Paragraphs.Add(CurrBlockStyle);
            else
                p = _doc.Body.Paragraphs.Add(text, CurrBlockStyle);

            if (ListStack.Count > 0)
            {
                var ls = ListStack.Peek();
                if (!ls.IsContinuation)
                {
                    p.ListFormat.Template = ls.Template;
                    if (ls.RestartAt >= 0)
                    {
                        p.RestartList();
                        // RestartList() creates and assigns a new list template to the paragraph:
                        ls.Template = p.ListFormat.Template;
                        ls.RestartAt = -1;
                    }
                }
                else
                {
                    p.ListFormat.Template = null;
                    p.Format.Indentation.LeftIndent = NumberedListContStyles[ListStack.Count - 1].ParagraphFormat.Indentation.LeftIndent;
                }
                p.ListFormat.LevelNumber = ls.Level;
                ls.IsContinuation = true;
                p.Format.Spacing.NoSpaceBetweenParagraphsOfSameStyle = true;
            }
            return p;
        }

        public Run AddRun(string text, Style style = null)
        {
            RunCollection runs;
            if (CurrHyperlink != null)
                runs = CurrHyperlink.GetRange().Runs;
            else
                runs = _doc.Body.Paragraphs.Last.GetRange().Runs;
            if (!string.IsNullOrEmpty(text) || runs.Last == null)
            {
                IsLastHtmlNewline = false;
                return runs.Add(text, style ?? CurrInlineStyle);
            }
            else
                return runs.Last;
        }

        public Break AddBreak()
        {
            IsLastHtmlNewline = false;
            return _doc.Body.Paragraphs.Last.GetRange().Runs.Last.GetRange().Texts.AddBreak();
        }

        public Hyperlink AddHyperlink()
        {
            return _doc.Body.Paragraphs.Last.GetRange().Hyperlinks.Add();
        }

        public Bookmark AddBookmark(string bookmark)
        {
            return _doc.Body.Paragraphs.Last.GetRange().Bookmarks.Add(bookmark);
        }

        /// <summary>
        /// Writes the specified content.
        /// </summary>
        /// <param name="content">The content.</param>
        /// <returns>This instance</returns>
        public WordRenderer Write(string content)
        {
            if (!string.IsNullOrEmpty(content))
            {
                AddRun(content);
            }
            InHtmlBlock = false;
            return this;
        }

        /// <summary>
        /// Writes the specified slice.
        /// </summary>
        /// <param name="slice">The slice.</param>
        /// <returns>This instance</returns>
        public WordRenderer Write(ref StringSlice slice)
        {
            Write(slice.AsSpan());
            return this;
        }

        /// <summary>
        /// Writes the specified slice.
        /// </summary>
        /// <param name="slice">The slice.</param>
        /// <returns>This instance</returns>
        public WordRenderer Write(StringSlice slice)
        {
            Write(slice.AsSpan());
            return this;
        }

        /// <summary>
        /// Writes the specified character.
        /// </summary>
        /// <param name="content">The content.</param>
        /// <returns>This instance</returns>
        public WordRenderer Write(char content)
        {
            if (content == '\n')
            {
                // previousWasLine = true;
                AddParagraph();
            }
            Write(content);
            return this;
        }

        /// <summary>
        /// Writes the specified content.
        /// </summary>
        /// <param name="content">The content.</param>
        /// <param name="offset">The offset.</param>
        /// <param name="length">The length.</param>
        /// <returns>This instance</returns>
        public WordRenderer Write(string content, int offset, int length)
        {
            if (content != null)
            {
                Write(content.AsSpan(offset, length));
            }
            return this;
        }

        /// <summary>
        /// Writes the specified content.
        /// </summary>
        /// <param name="content">The content.</param>
        public void Write(ReadOnlySpan<char> content, Style charStyle = null)
        {
            if (!content.IsEmpty)
            {
                AddRun(content.ToString(), charStyle);
            }
        }

        public WordRenderer WriteLine()
        {
            AddParagraph();
            // previousWasLine = true;
            return this;
        }

        public WordRenderer WriteLeafInline(LeafBlock leafBlock)
        {
            if (leafBlock is null)
                throw new ArgumentNullException(nameof(leafBlock));

            Inline inline = leafBlock.Inline;

            while (inline != null)
            {
                Write(inline);
                inline = inline.NextSibling;
            }

            return this;
        }

        public WordRenderer WriteCodeLines(CodeBlock codeBlock)
        {
            // TODO: optimize
            for (int i = 0; i < codeBlock.Lines.Count; ++i)
            {
                var txt = codeBlock.Lines.Lines[i].ToString();
                var tt = AddRun(txt);
                if (i < codeBlock.Lines.Count - 1)
                    tt.GetRange().Texts.AddBreak();
            }
            return this;
        }

        /// <summary>
        /// Gets a value to use as the base url for all relative links
        /// </summary>
        public Uri BaseUrl { get; set; }

        /// <summary>
        /// Allows links to be rewritten
        /// </summary>
        public Func<string, string> LinkRewriter { get; set; }

        /// <summary>
        /// Writes the lines of a <see cref="LeafBlock"/>
        /// </summary>
        /// <param name="leafBlock">The leaf block.</param>
        /// <param name="writeEndOfLines">if set to <c>true</c> write end of lines.</param>
        /// <returns>This instance</returns>
        public WordRenderer WriteLeafRawLines(LeafBlock leafBlock, bool writeEndOfLines)
        {
            if (leafBlock is null)
                throw new ArgumentNullException(nameof(leafBlock));

            var slices = leafBlock.Lines.Lines;
            if (slices != null)
            {
                for (int i = 0; i < slices.Length; i++)
                {
                    ref StringSlice slice = ref slices[i].Slice;
                    if (slice.Text is null)
                    {
                        break;
                    }
                    if (!writeEndOfLines && i > 0)
                    {
                        WriteLine();
                    }
                    if (_doc.Body.Paragraphs.Count == 0)
                        AddParagraph();
                    ReadOnlySpan<char> span = slice.AsSpan();
                    Write(span);
                    if (writeEndOfLines)
                    {
                        WriteLine();
                    }
                }
            }
            return this;
        }
    }
}
